// LICENSE: GPL

#include "EvieDaemon.h"

#include <X11/keysym.h>
#include <X11/Xutil.h>
#include "X11/Xatom.h"
#include <X11/Xlib.h>
#include <X11/X.h>
#include <X11/extensions/Xevie.h>
#include <X11/Intrinsic.h>


EvieDaemon::EvieDaemon( QObject* parent, unsigned int delay ) : QThread( parent )
{
  mTimer.setSingleShot( true );

  connect( &mTimer, SIGNAL( timeout() ), this, SLOT( onStopTyping() ) );

  connect( this, SIGNAL( startTimer() ), this, SLOT( onStartTyping() ) );

  mXevieRunning = false;
  mDisplay = NULL;

  setDelay( delay );

  setButtonEventsBlocked( false );
  setKeyEventsBlocked( false );

  setIgnoreModifiers( true );

  mDisplay = XOpenDisplay( NULL );

  initModifiers();

  if ( hasXevie() )
    grabXevie();
  else
  {
    qDebug( " " );
    qDebug( "Xevie initialization has FAILED! The reason might be " );
    qDebug( " " );
    qDebug( " * Xevie extension is disabled" );
    qDebug( "    Put a " );
    qDebug( "    Section \"Extensions\"" );
    qDebug( "      Option  \"XEVIE\" \"Enable\" " );
    qDebug( "    EndSection" );
    qDebug( "    in your /etc/X11/xorg.conf." );
    qDebug( " " );
    qDebug( " * Your X-Server is too old and has no support for Xevie" );
    qDebug( " " );
    qDebug( " * You are already running a program that uses the Xevie Extension" );
    qDebug( " " );
    qDebug( "Good luck!" );
  }
}

void EvieDaemon::onStartTyping()
{
  if ( !mTimer.isActive() )
    emit startTyping();

  mTimer.start();
}

void EvieDaemon::onStopTyping()
{
  if ( !hasKeyboardActivity() )
    emit stopTyping();
  else
    mTimer.start();
}

void EvieDaemon::start( QThread::Priority p )
{
  QThread::start( p );
}

bool EvieDaemon::grabXevie()
{
  if ( mXevieRunning || !mDisplay ) return false;

  XevieQueryVersion( mDisplay, &mVerMaj, &mVerMin );
  mXevieRunning = XevieStart( mDisplay );

  XevieSelectInput( mDisplay, 
                    KeyPressMask | ButtonPressMask | KeyReleaseMask | ButtonReleaseMask );

  start();
  qDebug( "Xevie %d.%d has successfully been started", majorVersionXevie(), minorVersionXevie() );

  return mXevieRunning;
}


void EvieDaemon::setDelay( unsigned int delay )
{
  unsigned int min = 10;
  unsigned int max = 2000;
  mTimer.setInterval( qBound( min, delay, max ) );
}

unsigned int EvieDaemon::delay() const
{
  return mTimer.interval();
}


void EvieDaemon::releaseXevie()
{
  if ( !mXevieRunning || !mDisplay ) return;

  terminate(); // FIXME kills thread, better by handling 'ClientMessage'
  XevieEnd( mDisplay );
  mXevieRunning = false;
}

EvieDaemon::~EvieDaemon()
{
  releaseXevie();

  if ( !mDisplay ) 
    XCloseDisplay( mDisplay );
}

void EvieDaemon::run()
{
  XEvent event;

  while ( mDisplay )
  {
    XNextEvent( mDisplay, &event );

    bool blocked = false;

    switch ( event.type )
    {
      case KeyPress:
      {
        XKeyEvent e = event.xkey;
        //qDebug( "key press %d, state %d", e.keycode, e.state ); 

        mKeyboardMap[ e.keycode ] = true;

        if ( !( isModifier( e.keycode ) && ignoreModifiers() ) )
        {
          emit keyPressed( e.keycode );
          emit startTimer();
        }

        blocked = keyEventsBlocked();

        break;
      }
      case KeyRelease:
      {
        XKeyEvent e = event.xkey;
        //qDebug( "key release %d, state %d", e.keycode, e.state ); 

        mKeyboardMap[ e.keycode ] = false;

        if ( !( isModifier( e.keycode ) && ignoreModifiers() ) )
          emit keyReleased( e.keycode );

        blocked = keyEventsBlocked();

        break;
      }
      case ButtonPress:
      {
        XButtonEvent e = event.xbutton;
        //qDebug( "button press %d, state %d", e.button, e.state ); 
        mButtonMap[ e.button ] = true;
        emit buttonPressed( e.button );

        blocked = buttonEventsBlocked();

        break;
      }
      case ButtonRelease:
      {
        XButtonEvent e = event.xbutton;
        //qDebug( "button release %d, state %d", e.button, e.state ); 
        mButtonMap[ e.button ] = false;
        emit buttonReleased( e.button );

        blocked = buttonEventsBlocked();

        break;
      }
      default:
        qDebug( "unknown" ); break;
    }

    if ( !blocked )
      XevieSendEvent( mDisplay, &event, XEVIE_UNMODIFIED );
  }

  qDebug( "exit!" );
}

bool EvieDaemon::hasKeyboardActivity() const
{
  QMap < int, bool > ::Iterator it;
  QMap < int, bool > m = mKeyboardMap;

  for ( it = m.begin(); it != m.end(); ++it )
  {
    if ( !( ignoreModifiers() && isModifier( it.key() ) ) )
      if ( it.value() ) return true;
  }

  return false;
}

bool EvieDaemon::hasButtonActivity() const
{
  QMap < int, bool > ::Iterator it;
  QMap < int, bool > m = mButtonMap;

  for ( it = m.begin(); it != m.end(); ++it ) 
  {
     if ( ! ( it.key() & Qt::MODIFIER_MASK ) )
       if ( it.value() ) return true;
  }

  return false;
}

int EvieDaemon::majorVersionXevie() const
{
  return mVerMaj;
}

int EvieDaemon::minorVersionXevie() const
{
  return mVerMin;
}

bool EvieDaemon::hasXevie() const
{
  if ( !mDisplay ) return false;

  int maj, min;

  XevieQueryVersion( mDisplay, &maj, &min );

  if ( maj == 0 && min == 0 )
    return false;
  else
  {
    if ( mXevieRunning ) return true;

    bool xevieRunnable = XevieStart( mDisplay );
    XevieEnd( mDisplay );

    return xevieRunnable;
  }
}

void EvieDaemon::initModifiers()
{
  if ( !mDisplay ) return;

  // read the current modifier map
  XModifierKeymap* mods = XGetModifierMapping( mDisplay );

  for ( int i = 0; i < 8 * mods->max_keypermod; i++ )
  {
    int mod = mods->modifiermap[i];

    if ( mModifierMap.indexOf( mod ) == -1 )
      mModifierMap.append( mods->modifiermap[i] );
  }

  XFreeModifiermap( mods );
}

bool EvieDaemon::ignoreModifiers() const
{
  return mIgnoreModifiers;
}

void EvieDaemon::setIgnoreModifiers( bool ignore )
{
  mIgnoreModifiers = ignore;
}

bool EvieDaemon::isModifier( int kc ) const
{
  return ( mModifierMap.indexOf( kc ) != -1 );
}

void EvieDaemon::setButtonEventsBlocked( bool blocked )
{
  mButtonEventsBlocked = blocked;
}

void EvieDaemon::setKeyEventsBlocked( bool blocked )
{
  mKeyEventsBlocked = blocked;
}

bool EvieDaemon::buttonEventsBlocked() const
{
  return mButtonEventsBlocked;
}

bool EvieDaemon::keyEventsBlocked() const
{
  return mKeyEventsBlocked;
}


