// LICENSE: GPL

#ifndef EVIEDAEMON_H
#define EVIEDAEMON_H

#include <QThread>
#include <QMap>
#include <QList>
#include <QtGui>
#include <QTimer>

#include <X11/Xlib.h> // FIXME

// TODO
//
// [*] call releaseXevie() before the process is being suspended
//     and grabXevie() it after wakeup
//     ATM all events got frozen on suspend which is nothing we want at all!!
//
// [*] make a sort-of singleton out of that class
//
// [*] make a lib out of it
//
// [*] support MIT_SCREEN_SAVER extension (for motion events)
//
// [*] find a proper way to end the event loop rather than terminate() it!
//
// [*] ...
//
// [*] fetch inital button map / keymap state on startup


class EvieDaemon : public QThread
{
    Q_OBJECT

    public:
        //! creates the Xevie daemon with a parent and a typing delay
        EvieDaemon( QObject* parent, unsigned int delay = 500 );

        //! destructor
        virtual ~EvieDaemon();

        //! sets the typing delay between key presses and releases
        void setDelay( unsigned int );

        //! returns the typing delay between key presses and releases
        unsigned int delay() const;

        //! checks whether there's keyboard activity
        bool hasKeyboardActivity() const;

        //! checks whether there's button activity
        bool hasButtonActivity() const;

        //! returns major version of Xevie
        int majorVersionXevie() const;

        //! returns minor version of Xevie
        int minorVersionXevie() const;

        //! checks whether Xevie is available
        //! FIXME seems currently broken which is X's fault!
        bool hasXevie() const;

        //! are modifier key events being ignored?
        bool ignoreModifiers() const;

        //! set whether modifier key events should be ignored
        void setIgnoreModifiers( bool );

        //! checks whether a key is a modifier
        bool isModifier( int ) const;

        //! set whether button events should be blocked
        void setButtonEventsBlocked( bool );

        //! set whether key events should be blocked
        void setKeyEventsBlocked( bool );

        //! checks whether button events are blocked currently
        bool buttonEventsBlocked() const;

        //! checks whether key events are blocked currently
        bool keyEventsBlocked() const;

    signals:
        //! emits pressed button (mouse/pad/...)
        //! regardless whether button events are being blocked or not
        void buttonPressed( int );

        //! emits released button (mouse/pad/...)
        //! regardless whether button events are being blocked or not
        void buttonReleased( int );

        //! emits pressed key
        //! regardless whether key events are being blocked or not 
        void keyPressed( int );

        //! emits released key
        //! regardless whether key events are being blocked or not
        void keyReleased( int );

        //! emitted when typing starts, ignores keyboard modifiers if set
        //! regardless whether key events are being blocked or not
        void startTyping();

        //! emitted after the last key release event
        //! regardless whether key events are being blocked or not
        void stopTyping();

    protected:
        //! protect the start routine
        void start( QThread::Priority = QThread::InheritPriority );

        //! event loop
        virtual void run();

        //! grabs Xevie (for receiving Xevie events)
        bool grabXevie();

        //! releases Xevie
        void releaseXevie();

        //! reads currect X modifiers
        void initModifiers();

    signals:
        //! starts the timer of the other thread
        void startTimer();

    protected slots:
        //! first key press event
        void onStartTyping();

        //! called after the delay has passed
        void onStopTyping();

    private:
        //! version numbers of Xevie
        int mVerMaj, mVerMin;

        //! display variable for Xlib calls
        Display* mDisplay; // FIXME

        //! keyboard map
        QMap < int, bool > mKeyboardMap;

        //! keyboad modifier list
        QList < int > mModifierMap;

        //! button map
        QMap < int, bool >  mButtonMap;

        //! is Xevie running?
        bool mXevieRunning;

        //! are modifiers being ignored?
        bool mIgnoreModifiers;

        //! button events are blocked?
        bool mButtonEventsBlocked;

        //! key events are blocked?
        bool mKeyEventsBlocked;

        //! 
        QTimer mTimer;
};

#endif

